/* jshint moz: true */

var self = require('sdk/self');

const __temp = require('chrome');
const Cc = __temp.Cc;
const Ci = __temp.Ci;
const Cr = __temp.Cr;

console.log("TEST	")

/*
This is the page for which we want to rewrite requests / responses.
*/
var targetPage = /^https?\:\/\/httpbin\.org\/*/;

var headerReplace = [{
  search: 'X-Frame-Options',
  replace: ''
}];

var contentReplace = [{
  search: /top\.location/,
  replace: 'location'
}];

var TracingListener = function TracingListener() {};
TracingListener.prototype = {
	/* The listener originally assigned to receive the HTTP request;
	 * Must pass request to original destination each time we intercept
	 * a callback.
	 */
	originalListener: null,

	/*
	 * Data to search and replace
	 */
	searchreplace: null,

	/*
	 * Called to signal the beginning of an asynchronous request
	 */
	onStartRequest: function (request, context) {

		// Initialize the staging area for this instance
		this.stagedData = [];

		/* Pass the onStartRequest call to the next listener in the chain */
		this.originalListener.onStartRequest(request, context);
	},

	/*
	 * Called when a chunk of data may be read without blocking the calling thread.
	 * - must always read exactly 'count' bytes from inputStream before returning
	 */
	onDataAvailable: function (request, context, inputStream, offset, count) {
		var binaryInputStream, storageStream, binaryOutputStream, segmentSizeInBytes,
		data, httpChannel, regexp_calendarscript, scriptLength,
		startIndex, lastIndex, oldScript, spaces, newScript, newScriptSpaces;

		// Create instances of the streams needed to read and pass data to the next
		// listener.  Read from the input stream, get a new input stream to pass
		// to the next listener from a storage stream, and output to the storage
		// stream by writing to the output stream.
		// Here, we create the instances for these streams:
		binaryInputStream = Cc["@mozilla.org/binaryinputstream;1"].createInstance(Ci.nsIBinaryInputStream);
		storageStream = Cc["@mozilla.org/storagestream;1"].createInstance(Ci.nsIStorageStream);
		binaryOutputStream = Cc["@mozilla.org/binaryoutputstream;1"].createInstance(Ci.nsIBinaryOutputStream);
		segmentSizeInBytes = 8192;

		// Bind the passed input stream to our instance
		binaryInputStream.setInputStream(inputStream);

		// Initialize the storage stream and bind the output stream to it
		storageStream.init(segmentSizeInBytes, count, null);
		binaryOutputStream.setOutputStream(storageStream.getOutputStream(0));

		// (REQUIRED) Read 'count' bytes from the input stream
		data = binaryInputStream.readBytes(count);

		// Fetch the HTTP channel interface to make queries on it
		httpChannel = request.QueryInterface(Ci.nsIHttpChannel);

		/*
		 * Search/replace data
		 */
		// match URI
		if((httpChannel.URI.spec.match(targetPage))) {
			// match data
			for(var j in contentReplace) {
				var matches = data.match(contentReplace[j].search);

				if(matches) {
					console.log('Matched ' + contentReplace[j].search);
					for(var m in matches) {
						if(m.match('[^0-9]')) continue;
						var rplc = contentReplace[j].replace;
						console.log('Replacing ' + matches[m] + ' with ' + rplc);
						if (rplc.length <= matches[m].length) {
							while (rplc.length < matches[m].length) rplc = rplc + ' '; // pad with spaces
							data = data.replace(matches[m], rplc);
						} else {
							// Replace string longer than search string -- can't proceed
						}
					}
				}
			}
		}

		// Write data to the output stream
		binaryOutputStream.writeBytes(data, count);

		// Pass the onDataAvailable call to the next listener in the chain
		this.originalListener.onDataAvailable(request, context, storageStream.newInputStream(0), offset, count);
	},

	/*
	 * Called to signal the end of an asynchronous request
	 * - always preceded by onStartRequest
	 */
	onStopRequest: function (request, context, statusCode) {

		/* Pass the onStopRequest call to the next listener in the chain */
		this.originalListener.onStopRequest(request, context, statusCode);
	},

	/*
	 * Satisfies nsISupports
	 */
	QueryInterface: function (aIID) {

		if (aIID.equals(Ci.nsIStreamListener) ||
		aIID.equals(Ci.nsISupports)) {
			return this;
		} else {
			throw Cr.NS_NOINTERFACE;
		}
	}
};

/*
 * Get a response header from an nsIHttpChannel, in Javascript semantics
 *
 * @param   {nsIHttpChannel} subject an HTTP response channel
 * @param   {string}     header An HTTP header (eg: "Set-Cookie" or "X-Frame-Options")
 * @returns {string} value of header; empty string (falsy) if not found or not available
 */
var getResponseHeader = function (subject, header) {
	var headerValue = "";
	try {
		// setResponseHeader( {string} header, {string} value, {boolean} merge )
		// if value is empty and merge is false, header will be cleared.
		// @throws NS_ERROR_NOT_AVAILABLE if called before received
		//     NS_ERROR_ILLEGAL_VALUE if changing the value of this response header
		//                is not allowed
		headerValue = subject.getResponseHeader(header);
		console.log('Value of header ' + header + ' is ' + headerValue);
	} catch (exception) {
		headerValue = "";
		console.log('Cannot get value of header ' + header);
	}
	return headerValue;
};

var myObserver = {
	/*
	 * nsIObserverService will call this function whenever an event is received
	 *
	 * This method will be subscribed to "http-on-examine-response".  It may
	 * be called multiple times for a single HTTP response as parts are received.
	 * One response usually corresponds to a single resource.
	 *
	 * This method implements the X-Frame-Options stripping, with the conditional
	 * logic implemented by isLocationI2().
	 *
	 * @public
	 * @methodOf httpMangler#
	 *
	 * @param {nsIHttpChannel} subject
	 *                 https://developer.mozilla.org/en/XPCOM_Interface_Reference/nsIHttpChannel
	 * @param {string}         topic
	 *                 http-on-examine-request (on recv response) or
	 *                 http-on-examine-cached-response (on read cached headers)
	 * @param              data
	 */
	observe: function (subject, topic, data) {

		// Change the get an {nsIHttpChannel} out of the {nsISupports} subject.
		var httpChannel = subject.QueryInterface(Ci.nsIHttpChannel);

		// The URL of the resource being requested
		var url = httpChannel.URI.spec;

		try {

			if(httpChannel.URI.spec.match(targetPage)) {
				console.log('Matched URI: ' + httpChannel.URI.spec);

				// Replace headers
				var headerValue;
				for(var i in headerReplace) {

					// Retrieve value of header
					headerValue = getResponseHeader(httpChannel, headerReplace[i].search);

					if(headerValue) {
						console.log('Replacing header: ' + headerReplace[i].search + ' with: ' + headerReplace[i].replace);
						try {
							httpChannel.setResponseHeader(headerReplace[i].search, headerReplace[i].replace, false);
						} catch (ex) {
							console.log('Failed to replace header ' + headerReplace[i].search + ' to ' + headerReplace[i].replace);
						}

					}
				}

				// Attach content search/replace
				try {
					var scrubber = new TracingListener();
					var tC = httpChannel.QueryInterface(Ci.nsITraceableChannel); // grab interface
					scrubber.originalListener = tC.setNewListener(scrubber); // insert into chain
				} catch (e) {
					console.log("setNewListener exception");
				}

			}
		} catch(e){
			console.log('Header observer exception');
		}
	}
};

/*
 * Implementing this method is part of the nsIObserver interface contract
 * Satisfies part of the nsIObserver interface
 *
 * @public
 * @methodOf IntegratedInbox.Http.httpMangler#
 *
 * @param {nsIIDRef} aIID An interface ID
 * @throws NS_NOINTERFACE Component does not have requested interface
 */
var QueryInterface = function (aIID) {
	if (aIID.equals(Ci.nsIObserver) ||
	aIID.equals(Ci.nsISupports)) {
		return this;
	} else {
		throw Cr.NS_NOINTERFACE;
	}
};

/*
 * Register to listen in on HTTP responses
 * @public
 * @methodOf IntegratedInbox.Http.httpMangler#
 */
var register = function () {

	// Retreive observer service to listen for HTTP responses
	var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);

	// Subscribe to http response observer
	observerService.addObserver(
		myObserver, // {object} Subscribing object with observer interface
		"http-on-examine-response", // {string} Topic ("channel") to subscribe to
		false
	); // {bool}   nsIWeakReference: allow object to die while registered

	observerService.addObserver(
		myObserver, // {object} Subscribing object with observer interface
		"http-on-examine-cached-response", // {string} Topic ("channel") to subscribe to
		false
	); // {bool}   nsIWeakReference: allow object to die while registered

	observerService.addObserver(
		myObserver, // {object} Subscribing object with observer interface
		"http-on-examine-merged-response", // {string} Topic ("channel") to subscribe to
		false
	); // {bool}   nsIWeakReference: allow object to die while registered

};

/*
 * Unregister to listen in on HTTP responses
 * @public
 * @methodOf IntegratedInbox.Http.httpMangler#
 */
var unregister = function () {
	var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
	observerService.removeObserver(this, http_response_topic);
};

// Off we go
register();
